"""Meta URL handler with automatic scheme-based switching of implementations"""

# allow for |-type UnionType declarations
from __future__ import annotations

from importlib import import_module
import json
import logging
from pathlib import Path
import re
from typing import Dict

from datalad_next.config import ConfigManager
from datalad_next.exceptions import CapturedException

from .base import UrlOperations

lgr = logging.getLogger('datalad.ext.next.url_operations.any')


__all__ = ['AnyUrlOperations']

# define handlers for each supported URL pattern
# FORMAT OF HANDLER REGISTRY (dict)
# - key: regex match expression to be apply on a URL (to test whether a
#   particular handler should be used for a given URL)
# - value: tuple (handler specification, see below)
# FORMAT OF HANDLER SPECIFICATION
# - tuple of min-length 1
# - item1: str, handler class to import
#   e.g., package.module.class
# - item2: dict, optional, kwargs to pass to the handler constructor

# TODO support proper entrypoint mechanism
# It is best to only record handlers here for which there is no alternative,
# because the best handler is determined based on this information
# and only this handler is imported. If that fails, there is no fallback.
# Handlers that may or may not work under given conditions should only
# be added via external logic after they have been found to be "working"
# on a given installation.
_url_handlers = {
    'http': ('datalad_next.url_operations.http.HttpUrlOperations',),
    'file': ('datalad_next.url_operations.file.FileUrlOperations',),
    'ssh': ('datalad_next.url_operations.ssh.SshUrlOperations',),
}


class AnyUrlOperations(UrlOperations):
    """Handler for operations on any supported URLs

    The methods inspect a given URL and call the corresponding
    methods for the `UrlOperations` implementation that matches the URL best.
    The "best match" is the match expression of a registered URL handler
    that yields the longest match against the given URL.

    Parameter identity and semantics are unchanged with respect to the
    underlying implementations. See their documentation for details.

    An instance retains and reuses URL scheme handler instances for subsequent
    operations, such that held connections or cached credentials can be reused
    efficiently.
    """
    def __init__(self, cfg: ConfigManager | None = None):
        """
        Parameters
        ----------
        cfg: ConfigManager, optional
          A config manager instance that is consulted for any configuration
          filesystem configuration individual handlers may support.
        """
        super().__init__(cfg=cfg)
        self._load_handler_registery()
        # cache of already used handlers
        self._url_handler_cache = dict()

    def _load_handler_registery(self):
        # update with handlers from config
        # https://github.com/datalad/datalad-next/issues/217
        cfgh = {}
        for citem in self.cfg.keys():
            if not citem.startswith('datalad.url-handler.'):
                # none of our business
                continue
            # the match expression is right in the item key
            # (all but the first two and the last segment)
            citem_l = citem.split('.')
            match = '.'.join(citem_l[2:-1])
            prop = citem_l[-1]
            value = self.cfg[citem]
            if prop != 'class':
                try:
                    value = json.loads(value)
                except Exception as e:
                    ce = CapturedException(e)
                    lgr.debug(
                        'Ignoring invalid URL handler configuration '
                        'for %r(%s): %r [%s]',
                        match, prop, value, ce)
                    continue
            hc = cfgh.get(match, {})
            hc[prop] = value
            cfgh[match] = hc
        # merge all specs
        uh = dict(_url_handlers)
        for match, spec in cfgh.items():
            try:
                uh[match] = (spec['class'], spec['kwargs'])
            except KeyError:
                try:
                    uh[match] = (spec['class'],)
                except Exception as e:
                    CapturedException(e)
                    lgr.debug(
                        'Ignoring incomplete URL handler specification '
                        'for %r: %r', match, spec)
        self._url_handlers = {}
        for k, v in uh.items():
            # compile matches to finalize
            lgr.log(8, 'Add URL handler for %r: %r', k, v)
            self._url_handlers[re.compile(k)] = v

    def _get_handler(self, url: str) -> UrlOperations:
        # match URL against all registered handlers and get the one with the
        # longest (AKA best) match
        longest_match = 0
        best_match = None
        for r in self._url_handlers:
            m = r.match(url)
            if not m:
                continue
            length = m.end() - m.start()
            if length > longest_match:
                best_match = r
                longest_match = length

        if best_match is None:
            raise ValueError(f'unsupported URL {url!r}')

        # reuse existing handler, they might already have an idea on
        # authentication etc. from a previously processed URL
        if best_match in self._url_handler_cache:
            return self._url_handler_cache[best_match]

        # we need to import the handler
        try:
            handler_spec = self._url_handlers[best_match]
            # split the import declaration into units
            toimport = handler_spec[0].split('.')
            # the handler class is the last unit
            cls = toimport[-1]
            # the rest is the module
            mod = '.'.join(toimport[:-1])
            module = import_module(mod, package='datalad')
            handler_cls = getattr(module, cls)
            handler_kwargs = handler_spec[1] if len(handler_spec) > 1 else {}
            url_handler = handler_cls(cfg=self.cfg, **handler_kwargs)
        except Exception as e:
            raise ValueError(
                'Cannot create URL handler instance for '
                f'{best_match.pattern!r} from {self._url_handlers[best_match]}') from e

        self._url_handler_cache[best_match] = url_handler
        return url_handler

    def is_supported_url(self, url) -> bool:
        return any(r.match(url) for r in self._url_handlers)

    def stat(self,
             url: str,
             *,
             credential: str | None = None,
             timeout: float | None = None) -> Dict:
        """Call `*UrlOperations.stat()` for the respective URL scheme"""
        return self._get_handler(url).stat(
            url, credential=credential, timeout=timeout)

    def download(self,
                 from_url: str,
                 to_path: Path | None,
                 *,
                 credential: str | None = None,
                 hash: list[str] | None = None,
                 timeout: float | None = None) -> Dict:
        """Call `*UrlOperations.download()` for the respective URL scheme"""
        return self._get_handler(from_url).download(
            from_url, to_path, credential=credential, hash=hash,
            timeout=timeout)

    def upload(self,
               from_path: Path | None,
               to_url: str,
               *,
               credential: str | None = None,
               hash: list[str] | None = None,
               timeout: float | None = None) -> Dict:
        """Call `*UrlOperations.upload()` for the respective URL scheme"""
        return self._get_handler(to_url).upload(
            from_path, to_url, credential=credential, hash=hash,
            timeout=timeout)

    def delete(self,
               url: str,
               *,
               credential: str | None = None,
               timeout: float | None = None) -> Dict:
        """Call `*UrlOperations.delete()` for the respective URL scheme"""
        return self._get_handler(url).delete(
            url, credential=credential, timeout=timeout)



import pytest

from datalad_next.exceptions import IncompleteResultsError
from datalad_next.tests import assert_result_count


def test_substitution_config_default(existing_dataset, no_result_rendering):
    ds = existing_dataset

    if ds.config.get('datalad.run.substitutions.python') is not None:
        # we want to test default handling when no config is set
        pytest.skip(
            'Test assumptions conflict with effective configuration')

    # the {python} placeholder is not explicitly defined, but it has
    # a default, which run() should discover and use
    res = ds.run('{python} -c "True"')
    assert_result_count(res, 1, action='run', status='ok')

    # make sure we could actually detect breakage with the check above
    with pytest.raises(IncompleteResultsError):
        ds.run('{python} -c "breakage"')



# emacs: -*- mode: python; py-indent-offset: 4; tab-width: 4; indent-tabs-mode: nil -*-
# ex: set sts=4 ts=4 sw=4 et:
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
#
#   See COPYING file distributed along with the datalad package for the
#   copyright and license terms.
#
# ## ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ### ##
"""High-level interface for creating a combi-target on a WebDAV capable server
 """
import logging
from urllib.parse import (
    quote as urlquote,
    urlunparse,
)

from datalad_next.commands import (
    EnsureCommandParameterization,
    ValidatedInterface,
    Parameter,
    build_doc,
    datasetmethod,
    eval_results,
    generic_result_renderer,
    get_status_dict,
)
from datalad.interface.common_opts import (
    recursion_flag,
    recursion_limit
)
from datalad_next.constraints import (
    ConstraintError,
    DatasetParameter,
    EnsureBool,
    EnsureChoice,
    EnsureDataset,
    EnsureInt,
    EnsureParsedURL,
    EnsureRange,
    EnsureRemoteName,
    EnsureStr,
    ParameterConstraintContext,
)
from datalad_next.utils import CredentialManager
from datalad_next.utils import (
    get_specialremote_credential_properties,
    patched_env,
    update_specialremote_credential,
    _yield_ds_w_matching_siblings,
)


__docformat__ = "restructuredtext"

lgr = logging.getLogger('datalad.distributed.create_sibling_webdav')


class CreateSiblingWebDAVParamValidator(EnsureCommandParameterization):
    def __init__(self):
        super().__init__(
            param_constraints=dict(
                url=EnsureParsedURL(
                    required=['scheme', 'netloc'],
                    forbidden=['query', 'fragment'],
                    match='^(http|https)://',
                ),
                dataset=EnsureDataset(
                    installed=True, purpose='create WebDAV sibling(s)'),
                name=EnsureRemoteName(),
                storage_name=EnsureRemoteName(),
                mode=EnsureChoice(
                    'annex', 'filetree', 'annex-only', 'filetree-only',
                    'git-only',
                ),
                # TODO https://github.com/datalad/datalad-next/issues/131
                credential=EnsureStr(),
                existing=EnsureChoice('skip', 'error', 'reconfigure'),
                recursive=EnsureBool(),
                recursion_limit=EnsureInt() & EnsureRange(min=0),
            ),
            validate_defaults=('dataset',),
            joint_constraints={
                ParameterConstraintContext(('url',), 'url'):
                self._validate_param_url,
                ParameterConstraintContext(
                    ('url', 'name'), 'default name'):
                self._validate_default_name,
                ParameterConstraintContext(
                    ('mode', 'name', 'storage_name'), 'default storage name'):
                self._validate_default_storage_name,
                ParameterConstraintContext(
                    ('mode', 'name', 'storage_name'), 'default storage name'):
                self._validate_default_storage_name,
                ParameterConstraintContext(
                    ('existing', 'recursive', 'name', 'storage_name',
                     'dataset', 'mode')):
                self._validate_existing_names,
            },
        )

    def _validate_param_url(self, url):
        if url.scheme == "http":
            lgr.warning(
                f"Using 'http:' ({url.geturl()!r}) means that WebDAV "
                "credentials are sent unencrypted over network links. "
                "Consider using 'https:'.")

    def _validate_default_name(self, url, name):
        if not name:
            # not using .netloc to avoid ports to show up in the name
            return {'name': url.hostname}

    def _validate_default_storage_name(self, mode, name, storage_name):
        if mode in ('annex-only', 'filetree-only') and storage_name:
            lgr.warning(
                "Sibling name will be used for storage sibling in "
                "storage-sibling-only mode, but a storage sibling name "
                "was provided"
            )
        if mode == 'git-only' and storage_name:
            lgr.warning(
                "Storage sibling setup disabled, but a storage sibling name "
                "was provided"
            )
        if mode != 'git-only' and not storage_name:
            storage_name = f"{name}-storage"

        if mode != 'git-only' and name == storage_name:
            # leads to unresolvable, circular dependency with publish-depends
            self.raise_for(
                dict(mode=mode, name=name, storage_name=storage_name),
                "sibling names must not be equal",
            )
        return dict(mode=mode, name=name, storage_name=storage_name)

    def _validate_existing_names(
            self, existing, recursive, name, storage_name, dataset,
            mode):
        if recursive:
            # we don't do additional validation for recursive processing,
            # this has to be done when things are running, because an
            # upfront validation would require an expensive traversal
            return

        if existing != 'error':
            # nothing to check here
            return

        if not isinstance(dataset, DatasetParameter):
            # we did not get a proper dataset parameter,
            # hence cannot tailor to a dataset to check a remote
            # name against
            return

        validator = EnsureRemoteName(known=False, dsarg=dataset)
        try:
            if mode != 'annex-only':
                validator(name)
            if mode != 'git-only':
                validator(storage_name)
        except ConstraintError as e:
            self.raise_for(
                dict(existing=existing,
                     recursive=recursive,
                     name=name,
                     storage_name=storage_name,
                     dataset=dataset,
                     mode=mode),
                e.msg,
            )
        return


@build_doc
class CreateSiblingWebDAV(ValidatedInterface):
    """Create a sibling(-tandem) on a WebDAV server

    WebDAV is a standard HTTP protocol extension for placing files on a server
    that is supported by a number of commercial storage services (e.g.
    4shared.com, box.com), but also instances of cloud-storage solutions like
    Nextcloud or ownCloud. These software packages are also the basis for
    some institutional or public cloud storage solutions, such as EUDAT B2DROP.

    For basic usage, only the URL with the desired dataset location on a WebDAV
    server needs to be specified for creating a sibling. However, the sibling
    setup can be flexibly customized (no storage sibling, or only a storage
    sibling, multi-version storage, or human-browsable single-version storage).

    This command does not check for conflicting content on the WebDAV
    server!

    When creating siblings recursively for a dataset hierarchy, subdataset
    exports are placed at their corresponding relative paths underneath the
    root location on the WebDAV server.


    Collaboration on WebDAV siblings

    The primary use case for WebDAV siblings is dataset deposition, where
    only one site is uploading dataset and file content updates.
    For collaborative workflows with multiple contributors, please make sure
    to consult the documentation on the underlying ``datalad-annex::``
    Git remote helper for advice on appropriate setups:
    http://docs.datalad.org/projects/next/


    Git-annex implementation details

    Storage siblings are presently configured to NOT be enabled
    automatically on cloning a dataset. Due to a limitation of git-annex, this
    would initially fail (missing credentials). Instead, an explicit
    ``datalad siblings enable --name <storage-sibling-name>`` command must be
    executed after cloning. If necessary, it will prompt for credentials.

    This command does not (and likely will not) support embedding credentials
    in the repository (see ``embedcreds`` option of the git-annex ``webdav``
    special remote; https://git-annex.branchable.com/special_remotes/webdav),
    because such credential copies would need to be updated, whenever they
    change or expire. Instead, credentials are retrieved from DataLad's
    credential system. In many cases, credentials are determined automatically,
    based on the HTTP authentication realm identified by a WebDAV server.

    This command does not support setting up encrypted remotes (yet). Neither
    for the storage sibling, nor for the regular Git-remote. However, adding
    support for it is primarily a matter of extending the API of this command,
    and passing the respective options on to the underlying git-annex setup.

    This command does not support setting up chunking for webdav storage
    siblings (https://git-annex.branchable.com/chunking).
    """
    _examples_ = [
       dict(text="Create a WebDAV sibling tandem for storage of a dataset's "
                 "file content and revision history. A user will be prompted "
                 "for any required credentials, if they are not yet known.",
             code_py="create_sibling_webdav(url='https://webdav.example.com/myds')",
             code_cmd='datalad create-sibling-webdav "https://webdav.example.com/myds"'),
       dict(text="Such a dataset can be cloned by DataLad via a specially "
                 "crafted URL. Again, credentials are automatically "
                 "determined, or a user is prompted to enter them",
            code_py="clone('datalad-annex::?type=webdav&encryption=none&url=https://webdav.example.com/myds')",
            code_cmd='datalad clone "datalad-annex::?type=webdav&encryption=none&url=https://webdav.example.com/myds"'),
       dict(
           text="A sibling can also be created with a human-readable file "
                 "tree, suitable for data exchange with non-DataLad users, "
                 "but only able to host a single version of each file",
           code_py="create_sibling_webdav(url='https://example.com/browseable', mode='filetree')",
           code_cmd='datalad create-sibling-webdav --mode filetree "https://example.com/browsable"'),
       dict(text="Cloning such dataset siblings is possible via a convenience "
                 "URL",
            code_py="clone('webdavs://example.com/browseable')",
            code_cmd='datalad clone "webdavs://example.com/browsable"'),
       dict(text="In all cases, the storage sibling needs to explicitly "
                 "enabled prior to file content retrieval",
            code_py="siblings('enable', name='example.com-storage')",
            code_cmd='datalad siblings enable --name example.com-storage'),
    ]

    _params_ = dict(
        url=Parameter(
            args=("url",),
            metavar='URL',
            doc="URL identifying the sibling root on the target WebDAV server"),
        dataset=Parameter(
            args=("-d", "--dataset"),
            doc="""specify the dataset to process.  If
            no dataset is given, an attempt is made to identify the dataset
            based on the current working directory"""),
        name=Parameter(
            args=('-s', '--name',),
            metavar='NAME',
            doc="""name of the sibling. If none is given, the hostname-part
            of the WebDAV URL will be used.
            With `recursive`, the same name will be used to label all
            the subdatasets' siblings."""),
        storage_name=Parameter(
            args=("--storage-name",),
            metavar="NAME",
            doc="""name of the storage sibling (git-annex special remote).
            Must not be identical to the sibling name. If not specified,
            defaults to the sibling name plus '-storage' suffix. If only
            a storage sibling is created, this setting is ignored, and
            the primary sibling name is used."""),
        credential=Parameter(
            args=("--credential",),
            metavar='NAME',
            doc="""name of the credential providing a user/password credential
            to be used for authorization. The credential can be supplied via
            configuration setting 'datalad.credential.<name>.user|secret', or
            environment variable DATALAD_CREDENTIAL_<NAME>_USER|SECRET, or will
            be queried from the active credential store using the provided
            name. If none is provided, the last-used credential for the
            authentication realm associated with the WebDAV URL will be used.
            Only if a credential name was given, it will be encoded in the
            URL of the created WebDAV Git remote, credential auto-discovery
            will be performed on each remote access.""",
        ),
        existing=Parameter(
            args=("--existing",),
            doc="""action to perform, if a (storage) sibling is already
            configured under the given name.
            In this case, sibling creation can be skipped ('skip') or the
            sibling (re-)configured ('reconfigure') in the dataset, or the
            command be instructed to fail ('error').""", ),
        recursive=recursion_flag,
        recursion_limit=recursion_limit,
        mode=Parameter(
            args=("--mode",),
            doc="""Siblings can be created in various modes:
            full-featured sibling tandem, one for a dataset's Git history
            and one storage sibling to host any number of file versions
            ('annex').
            A single sibling for the Git history only ('git-only').
            A single annex sibling for multi-version file storage only
            ('annex-only').
            As an alternative to the standard (annex) storage sibling setup
            that is capable of storing any number of historical file versions
            using a content hash layout ('annex'|'annex-only'), the 'filetree'
            mode can used.
            This mode offers a human-readable data organization on the WebDAV
            remote that matches the file tree of a dataset (branch).
            However, it can, consequently, only store a single version of each
            file in the file tree.
            This mode is useful for depositing a single dataset
            snapshot for consumption without DataLad. The 'filetree' mode
            nevertheless allows for cloning such a single-version dataset,
            because the full dataset history can still be pushed to the WebDAV
            server.
            Git history hosting can also be turned off for this setup
            ('filetree-only').
            When both a storage sibling and a regular sibling are created
            together, a publication dependency on the storage sibling is
            configured for the regular sibling in the local dataset clone.
            """),
    )

    _validator_ = CreateSiblingWebDAVParamValidator()

    @staticmethod
    @datasetmethod(name='create_sibling_webdav')
    @eval_results
    def __call__(
            url,
            *,
            dataset=None,
            name=None,
            storage_name=None,
            mode='annex',
            credential=None,
            existing='error',
            recursive=False,
            recursion_limit=None):

        ds = dataset.ds

        res_kwargs = dict(
            action='create_sibling_webdav',
            logger=lgr,
            refds=ds.path,
        )

        # Query existing siblings upfront in order to fail early on
        # existing=='error', since misconfiguration (particularly of special
        # remotes) only to fail in a subdataset later on with that config, can
        # be quite painful.
        if existing == 'error':
            # even if we have to fail, let's report all conflicting siblings
            # in subdatasets, an outside controller can stop the generator
            # if desired
            failed = False
            for dpath, sname in _yield_ds_w_matching_siblings(
                    ds,
                    (name, storage_name),
                    recursive=recursive,
                    recursion_limit=recursion_limit):
                res = get_status_dict(
                    status='error',
                    message=(
                        "a sibling %r is already configured in dataset %r",
                        sname, dpath),
                    type='sibling',
                    name=sname,
                    ds=ds,
                    **res_kwargs,
                )
                failed = True
                yield res
            if failed:
                return

        # determine the credential upfront
        # can be done once at the start, all siblings will live on the same
        # server
        # if all goes well, we'll store a credential (update) at the very end
        credman = CredentialManager(ds.config)
        credname, credprops = credman.obtain(
            credential,
            prompt='User name and password are required for WebDAV access '
                   f'at {url.geturl()}',
            query_props=get_specialremote_credential_properties(
                dict(type='webdav', url=url.geturl())),
            type_hint='user_password',
            # make it raise ValueError when the critical components are missing
            expected_props=['user', 'secret'],
        )
        cred_user = credprops['user']
        cred_password = credprops['secret']

        def _dummy(ds, refds, **kwargs):
            """Small helper to prepare the actual call to _create_sibling_webdav()
            for a particular (sub)dataset.

            We only have kwargs to catch whatever it throws at us.
            """
            relpath = ds.pathobj.relative_to(refds.pathobj) if not ds == refds else None
            if relpath:
                dsurl = f"{urlunparse(url)}/{relpath}"
            else:
                dsurl = url.geturl()

            return _create_sibling_webdav(
                ds,
                dsurl,
                # we pass the given, not the discovered, credential name!
                # given a name means "take this particular one", not giving a
                # name means "take what is best". Only if we pass this
                # information on, we achieve maintaining this behavior
                credential_name=credential,
                credential=(cred_user, cred_password),
                mode=mode,
                name=name,
                storage_name=storage_name,
                existing=existing,
            )

        # Generate a sibling for dataset "ds", and for sub-datasets if recursive
        # is True.
        for res in ds.foreach_dataset(
                _dummy,
                return_type='generator',
                result_renderer='disabled',
                recursive=recursive,
                # recursive False is not enough to disable recursion
                # https://github.com/datalad/datalad/issues/6659
                recursion_limit=0 if not recursive else recursion_limit,
        ):
            # unwind result generator
            for partial_result in res.get('result', []):
                yield dict(res_kwargs, **partial_result)

        # this went well, update the credential
        update_specialremote_credential(
            'webdav',
            credman,
            credname,
            credprops,
            credtype_hint='user_password',
            duplicate_hint=
            'Specify a credential name via the `credential` parameter '
            ' and/or configure a credential with the datalad-credentials '
            'command{}'.format(
                f' with a `realm={credprops["realm"]}` property'
                if 'realm' in credprops else ''),
        )

    @staticmethod
    def custom_result_renderer(res, **kwargs):
        from datalad_next.uis import (
            ansi_colors as ac,
            ui_switcher as ui,
        )
        from os.path import relpath

        if res['status'] != 'ok' or 'sibling_webdav' not in res['action'] or \
                res['type'] != 'sibling':
            # It's either 'notneeded' (not rendered), an `error`/`impossible` or
            # something unspecific to this command. No special rendering
            # needed.
            generic_result_renderer(res)
            return

        ui.message('{action}({status}): {path} [{name}{url}]'.format(
            action=ac.color_word(res['action'], ac.BOLD),
            path=relpath(res['path'], res['refds'])
            if 'refds' in res else res['path'],
            name=ac.color_word(res.get('name', ''), ac.MAGENTA),
            url=f": {res['url']}" if 'url' in res else '',
            status=ac.color_status(res['status']),
        ))


def _create_sibling_webdav(
        ds, url, *,
        credential_name, credential,
        mode='git-only', name=None, storage_name=None, existing='error'):
    """
    Parameters
    ----------
    ds: Dataset
    url: str
    credential_name: str
    credential: tuple
    mode: str, optional
    name: str, optional
    storage_name: str, optional
    existing: str, optional
    """
    # simplify downstream logic, export yes or no
    export_storage = 'filetree' in mode

    existing_siblings = [
        r[1] for r in _yield_ds_w_matching_siblings(
            ds,
            (name, storage_name),
            recursive=False)
    ]

    if mode != 'git-only':
        yield from _create_storage_sibling(
            ds,
            url,
            storage_name,
            credential,
            export=export_storage,
            existing=existing,
            known=storage_name in existing_siblings,
        )

    if mode not in ('annex-only', 'filetree-only'):
        yield from _create_git_sibling(
            ds,
            url,
            name,
            credential_name,
            credential,
            export=export_storage,
            existing=existing,
            known=name in existing_siblings,
            publish_depends=storage_name if mode != 'git-only'
            else None
        )


def _get_skip_sibling_result(name, ds, type_):
    return get_status_dict(
        action='create_sibling_webdav{}'.format(
            '.storage' if type_ == 'storage' else ''),
        ds=ds,
        status='notneeded',
        message=("skipped creating %r sibling %r, already exists",
                 type_, name),
        name=name,
        type='sibling',
    )


def _create_git_sibling(
        ds, url, name, credential_name, credential, export, existing,
        known, publish_depends=None):
    """
    Parameters
    ----------
    ds: Dataset
    url: str
    name: str
    credential_name: str
    credential: tuple
    export: bool
    existing: {skip, error, reconfigure}
    known: bool
        Flag whether the sibling is a known remote (no implied
        necessary existence of content on the remote).
    publish_depends: str or None
        publication dependency to set
    """
    if known and existing == 'skip':
        yield _get_skip_sibling_result(name, ds, 'git')
        return

    remote_url = \
        "datalad-annex::?type=webdav&encryption=none" \
        "&exporttree={export}&url={url}".format(
            export='yes' if export else 'no',
            # urlquote, because it goes into the query part of another URL
            url=urlquote(url))
    if credential_name:
        # we need to quote the credential name too.
        # e.g., it is not uncommon for credentials to be named after URLs
        remote_url += f'&dlacredential={urlquote(credential_name)}'

    # announce the sibling to not have an annex (we have a dedicated
    # storage sibling for that) to avoid needless annex-related processing
    # and speculative whining by `siblings()`
    ds.config.set(f'remote.{name}.annex-ignore', 'true', scope='local')

    for r in ds.siblings(
            # action must always be 'configure' (not 'add'), because above we just
            # made a remote {name} known, which is detected by `sibling()`. Any
            # conflict detection must have taken place separately before this call
            # https://github.com/datalad/datalad/issues/6649
            action="configure",
            name=name,
            url=remote_url,
            # this is presently the default, but it may change
            fetch=False,
            publish_depends=publish_depends,
            return_type='generator',
            result_renderer='disabled'):
        if r.get('action') == 'configure-sibling':
            r['action'] = 'reconfigure_sibling_webdav' \
                if known and existing == 'reconfigure' \
                else 'create_sibling_webdav'
        yield r


def _create_storage_sibling(
        ds, url, name, credential, export, existing, known=False):
    """
    Parameters
    ----------
    ds: Dataset
    url: str
    name: str
    credential: tuple
    export: bool
    existing: {skip, error, reconfigure}
        (Presently unused)
    known: bool
        Flag whether the sibling is a known remote (no implied
        necessary existence of content on the remote).
    """
    if known and existing == 'skip':
        yield _get_skip_sibling_result(name, ds, 'storage')
        return

    cmd_args = [
        'enableremote' if known and existing == 'reconfigure'
        else 'initremote',
        name,
        "type=webdav",
        f"url={url}",
        f"exporttree={'yes' if export else 'no'}",
        "encryption=none",
        # for now, no autoenable. It would result in unconditional
        # error messages on clone
        #https://github.com/datalad/datalad/issues/6634
        #"autoenable=true"
    ]
    # Add a git-annex webdav special remote. This requires to set
    # the webdav environment variables accordingly.
    with patched_env(WEBDAV_USERNAME=credential[0],
                     WEBDAV_PASSWORD=credential[1],
    ):
        ds.repo.call_annex(cmd_args)
    yield get_status_dict(
        ds=ds,
        status='ok',
        action='reconfigure_sibling_webdav.storage'
               if known and existing == 'reconfigure' else
        'create_sibling_webdav.storage',
        name=name,
        type='sibling',
        url=url,
    )



from inspect import isgenerator
from io import StringIO
import pytest
from tempfile import NamedTemporaryFile
from unittest.mock import patch
from pathlib import Path

from datalad_next.consts import on_windows
from datalad_next.exceptions import CapturedException

from ..base import DatasetParameter

from ..basic import (
    EnsureInt,
    EnsureBool,
    EnsurePath,
)
from ..compound import (
    ConstraintWithPassthrough,
    EnsureIterableOf,
    EnsureListOf,
    EnsureTupleOf,
    EnsureMapping,
    EnsureGeneratorFromFileLike,
    WithDescription,
)
from ..exceptions import ConstraintError


# imported from ancient test code in datalad-core,
# main test is test_EnsureIterableOf
def test_EnsureTupleOf():
    c = EnsureTupleOf(str)
    assert c(['a', 'b']) == ('a', 'b')
    assert c(['a1', 'b2']) == ('a1', 'b2')
    assert c.short_description() == "tuple(<class 'str'>)"


# imported from ancient test code in datalad-core,
# main test is test_EnsureIterableOf
def test_EnsureListOf():
    c = EnsureListOf(str)
    assert c(['a', 'b']) == ['a', 'b']
    assert c(['a1', 'b2']) == ['a1', 'b2']
    assert c.short_description() == "list(<class 'str'>)"
    assert repr(c) == \
        "EnsureListOf(item_constraint=<class 'str'>, min_len=None, max_len=None)"


def test_EnsureIterableOf():
    c = EnsureIterableOf(list, int)
    assert c.short_description() == "<class 'list'>(<class 'int'>)"
    assert c.item_constraint == int
    # testing aspects that are not covered by test_EnsureListOf
    tgt = [True, False, True]
    assert EnsureIterableOf(list, bool)((1, 0, 1)) == tgt
    assert EnsureIterableOf(list, bool, min_len=3, max_len=3)((1, 0, 1)) == tgt
    with pytest.raises(ValueError):
        # too many items
        EnsureIterableOf(list, bool, max_len=2)((1, 0, 1))
    with pytest.raises(ValueError):
        # too few items
        EnsureIterableOf(list, bool, min_len=4)((1, 0, 1))
    with pytest.raises(ValueError):
        # invalid specification min>max
        EnsureIterableOf(list, bool, min_len=1, max_len=0)
    with pytest.raises(ValueError):
        # item_constraint fails
        EnsureIterableOf(list, dict)([5.6, 3.2])
    with pytest.raises(ValueError):
        # item_constraint fails
        EnsureIterableOf(list, EnsureBool())([5.6, 3.2])

    seq = [3.3, 1, 2.6]

    def _mygen():
        for i in seq:
            yield i

    def _myiter(iter):
        for i in iter:
            yield i

    # feeding a generator into EnsureIterableOf and getting one out
    assert list(EnsureIterableOf(_myiter, int)(_mygen())) == [3, 1, 2]


def test_EnsureMapping(dataset):
    true_key = 5
    true_value = False

    constraint = EnsureMapping(EnsureInt(), EnsureBool(), delimiter='::')
    # called without a mapping type
    with pytest.raises(ValueError):
        constraint(true_key)

    assert 'mapping of int -> bool' in constraint.short_description()
    assert repr(constraint) == \
        "EnsureMapping(key=EnsureInt(), value=EnsureBool(), delimiter='::')"

    # must all work
    for v in ('5::no',
              [5, 'false'],
              ('5', False),
              {'5': 'False'},
    ):
        d = constraint(v)
        assert isinstance(d, dict)
        assert len(d) == 1
        k, v = d.popitem()
        assert k == true_key
        assert v == true_value

    # must all fail
    for v in ('5',
              [],
              tuple(),
              {},
              # additional value
              [5, False, False],
              {'5': 'False', '6': True}):
        with pytest.raises(ValueError):
            d = constraint(v)

    # test for_dataset()
    # smoketest
    ds = dataset
    cds = constraint.for_dataset(ds)
    assert cds._key_constraint == constraint._key_constraint.for_dataset(ds)
    assert cds._value_constraint == \
        constraint._value_constraint.for_dataset(ds)
    # test that the path is resolved for the dataset
    pathconstraint = \
        EnsureMapping(key=EnsurePath(), value=EnsureInt()).for_dataset(
            DatasetParameter(ds.pathobj, ds))
    assert pathconstraint('some:5') == {(Path.cwd() / 'some'): 5}
    pathconstraint = \
        EnsureMapping(key=EnsurePath(), value=EnsurePath()).for_dataset(
            DatasetParameter(ds, ds))
    assert pathconstraint('some:other') == \
           {(ds.pathobj / 'some'): (ds.pathobj / 'other')}


def test_EnsureGeneratorFromFileLike():
    item_constraint = EnsureMapping(EnsureInt(), EnsureBool(), delimiter='::')
    constraint = EnsureGeneratorFromFileLike(item_constraint)

    assert 'items of type "mapping of int -> bool" read from a file-like' \
        == constraint.short_description()
    assert repr(constraint) == \
        "EnsureGeneratorFromFileLike(" \
        "item_constraint=EnsureMapping(key=EnsureInt(), " \
        "value=EnsureBool(), delimiter='::'))"

    c = constraint(StringIO("5::yes\n1234::no\n"))
    assert isgenerator(c)
    assert list(c) == [{5: True}, {1234: False}]

    # missing final newline is not a problem
    c = constraint(StringIO("5::yes\n1234::no"))
    assert list(c) == [{5: True}, {1234: False}]

    # item constraint violation
    invalid_input = StringIO("1234::BANG\n5::yes")
    # immediate raise is default
    with pytest.raises(ValueError) as e:
        list(constraint(invalid_input))
    assert 'be convertible to boolean' in str(e)
    # but optionally it yields the exception to be able to
    # continue and enable a caller to raise/report/ignore
    # (must redefine `invalid_input` to read from start)
    invalid_input = StringIO("1234::BANG\n5::yes")
    res = list(
        EnsureGeneratorFromFileLike(
            item_constraint,
            exc_mode='yield',
        )(invalid_input)
    )
    # we get the result after the exception occurred
    assert isinstance(res[0], CapturedException)
    assert res[1] == {5: True}

    # read from STDIN
    with patch("sys.stdin", StringIO("5::yes\n1234::no")):
        assert list(constraint('-')) == [{5: True}, {1234: False}]

    with patch("sys.stdin", StringIO("5::yes\n1234::no")):
        # will unpack a length-1 sequence for convenience
        assert list(constraint(['-'])) == [{5: True}, {1234: False}]

    # read from file
    if not on_windows:
        # on windows the write-rewind-test logic is not possible
        # (PermissionError) -- too lazy to implement a workaround
        with NamedTemporaryFile('w+') as f:
            f.write("5::yes\n1234::no")
            f.seek(0)
            assert list(constraint(f.name)) == [{5: True}, {1234: False}]

    # invalid file
    with pytest.raises(ValueError) as e:
        list(constraint('pytestNOTHEREdatalad'))


def test_ConstraintWithPassthrough(dataset):
    wrapped = EnsureInt()
    cwp = ConstraintWithPassthrough(wrapped, passthrough='mike')
    # main purpose
    assert cwp('mike') == 'mike'
    assert cwp('5') == 5
    # most info is coming straight from `wrapped`, the pass-through is
    # meant to be transparent
    assert str(cwp) == str(wrapped)
    assert cwp.short_description() == wrapped.short_description()
    assert cwp.long_description() == wrapped.long_description()
    # but repr reveals it
    assert repr(cwp).startswith('ConstraintWithPassthrough(')
    # tailoring for a dataset keeps the pass-through
    ds = dataset
    cwp_ds = cwp.for_dataset(ds)
    assert cwp_ds.passthrough == cwp.passthrough
    assert cwp.constraint == wrapped.for_dataset(ds)


def test_WithDescription(dataset):
    wrapped = EnsureInt()
    # confirm starting point
    assert wrapped.input_synopsis == 'int'
    assert wrapped.input_description \
        == "value must be convertible to type 'int'"
    # we are actually not replacing anything
    c = WithDescription(wrapped)
    assert c.input_synopsis == wrapped.input_synopsis
    assert c.input_description == wrapped.input_description
    # with no dataset docs, the wrapping is removed on tailoring
    ds = dataset
    assert isinstance(
        c.for_dataset(DatasetParameter(None, ds)),
        EnsureInt)
    # check all replacements are working
    c = WithDescription(
        wrapped,
        input_synopsis='mysynopsis',
        input_description='mydescription',
        input_synopsis_for_ds='dssynopsis',
        input_description_for_ds='dsdescription',
        error_message='myerror',
        error_message_for_ds='dserror',
    )
    # function is maintained
    assert c('5') == 5
    assert str(c) == '<EnsureInt with custom description>'
    assert repr(c) == \
        "WithDescription(EnsureInt(), " \
        "input_synopsis='mysynopsis', " \
        "input_description='mydescription', " \
        "input_synopsis_for_ds='dssynopsis', " \
        "input_description_for_ds='dsdescription', " \
        "error_message='myerror', " \
        "error_message_for_ds='dserror')"
    assert c.constraint is wrapped
    assert c.input_synopsis == 'mysynopsis'
    assert c.input_description == 'mydescription'
    # description propagates through tailoring
    cds = c.for_dataset(DatasetParameter(None, ds))
    assert isinstance(cds, WithDescription)
    assert cds.input_synopsis == 'dssynopsis'
    assert cds.input_description == 'dsdescription'

    # when the wrapped constraint raises, the wrapper
    # interjects and reports a different error
    with pytest.raises(ConstraintError) as e:
        c(None)
    assert e.value.msg == 'myerror'

    # legacy functionality
    c.short_description() == c.input_synopsis
    c.long_description() == c.input_description



from pathlib import (
    Path,
    PurePosixPath,
)

from ..add_method_url2transport_path import (
    local_io_url2transport_path,
    http_remote_io_url2transport_path,
)

from datalad.utils import on_windows
from datalad.tests.utils_pytest import skip_if


@skip_if(on_windows)
def test_local_io_url2transport_path_posix():

    for url, transport_path in (
            ('/a/b/c', '/a/b/c'),
            ('/C:/a/b/c', '/C:/a/b/c'),
            ('C:/a/b/c', 'C:/a/b/c'),
    ):
        assert local_io_url2transport_path(
            None,
            PurePosixPath(url)
        ) == Path(transport_path)


@skip_if(not on_windows)
def test_local_io_url2transport_path_windows(monkeypatch):
    monkeypatch.setattr(
        'datalad_next.patches.add_method_url2transport_path.on_windows',
        True,
    )
    warnings = []
    monkeypatch.setattr(
        'datalad_next.patches.add_method_url2transport_path.lgr.warning',
        lambda x: warnings.append(x),
    )

    for url, transport_path in (
            ('/a/b/c', '/a/b/c'),
            ('C:/a/b/c', 'C:/a/b/c'),
            ('C:a/b/c', 'C:a/b/c'),
            ('/C:a/b/c', 'C:a/b/c'),
            ('/C:/a/b/c', 'C:/a/b/c'),
    ):
        assert local_io_url2transport_path(
            None,
            PurePosixPath(url)
        ) == Path(transport_path)
    assert len(warnings) > 0


def test_http_remote_io_url2transport_path():

    for url in ('a/b/c', '/a/b/c', '/C:/a/b/c', '/C:a/b/c', 'C:/a/b/c'):
        assert http_remote_io_url2transport_path(
            None,
            PurePosixPath(url)
        ) == PurePosixPath(url)



from . import (
    GeneratorMixIn,
    NoCapture,
    StdOutCapture,
)


#
# Below are generic generator protocols that should be provided
# upstream
#
class NoCaptureGeneratorProtocol(NoCapture, GeneratorMixIn):
    def __init__(self, done_future=None, encoding=None):
        NoCapture.__init__(self, done_future, encoding)
        GeneratorMixIn.__init__(self)

    def timeout(self, fd):
        raise TimeoutError(f"Runner timeout: process has not terminated yet")


class StdOutCaptureGeneratorProtocol(StdOutCapture, GeneratorMixIn):
    def __init__(self, done_future=None, encoding=None):
        StdOutCapture.__init__(self, done_future, encoding)
        GeneratorMixIn.__init__(self)

    def pipe_data_received(self, fd: int, data: bytes):
        assert fd == 1
        self.send_result(data)

    def timeout(self, fd):
        raise TimeoutError(f"Runner timeout {fd}")



"""``DatasetParameter`` support for ``resolve_path()``

This is the standard result of ``EnsureDataset``, which unlike
the datalad-core version actually carries a ``Dataset`` instance.

This patch ensure the traditional handling of "dataset instance from
a string-type parameter in this context.
"""

import logging

from . import apply_patch

# use same logger as -core, looks weird but is correct
lgr = logging.getLogger('datalad.dataset')


def resolve_path(path, ds=None, ds_resolved=None):
    if hasattr(ds, 'auto_instance_from_path'):
        # this instance came from datalad-next's EnsureDataset,
        # pretend that we resolved the dataset by hand
        return orig_resolve_path(
            ds=ds.auto_instance_from_path,
            ds_resolved=ds,
        )
    else:
        return orig_resolve_path(ds=ds, ds_resolved=ds_resolved)


# we need to preserve it as the workhorse, this patch only wraps around it
orig_resolve_path = apply_patch(
    'datalad.distribution.dataset', None, 'resolve_path',
    resolve_path,
    msg='Apply datalad-next patch to distribution.dataset:resolve_path')

# reuse docs
resolve_path.__doc__ = orig_resolve_path.__doc__



from datalad_next.utils import chpwd
from datalad_next.tests import run_main


def test_cli_configoverrides(existing_dataset):
    # test whether a `datalad -c ...` is effective within the
    # execution environment of a subprocess (for a non-datalad
    # configuration item
    with chpwd(existing_dataset.path):
        out, err = run_main(
            [
                '-c', 'bogusdataladtestsec.subsec=unique',
                'run',
                'git config bogusdataladtestsec.subsec',
            ],
            # git-config would fail, if the config item is unknown
            exit_code=0,
        )



"""Provide proper arguments for scp-command calls in `SSHConnection`

The original code has errors in the methods ``BaseSSHConnection.put``
``BaseSSHConnection.get``. Both methods use ``self.sshri.hostname`` to
determine the target for an ``scp``-command. They should instead use
``self.sshri.as_str()`` in order to include a user specification into the
target.

The changes in this patch use ``self.sshri.as_str()`` to provide the correct
targets for ``scp``-commands.
"""

import logging

from datalad.support.sshconnector import (
    StdOutErrCapture,
    ensure_list,
)
from datalad_next.patches import apply_patch


# use same logger as -core
lgr = logging.getLogger('datalad.support.sshconnector')


# The method 'BaseSSHConnection_get' is a patched version of
# 'datalad/support/sshconnector.py:BaseSSHConnection.get'
# from datalad@e0b357d9b8ca5f432638c23c0cb7c373028c8e52
def BaseSSHConnection_get(self, source, destination, recursive=False, preserve_attrs=False):
    """Copies source file/folder from remote to a local destination.

    Note: this method performs escaping of filenames to an extent that
    moderately weird ones should work (spaces, quotes, pipes, other
    characters with special shell meaning), but more complicated cases
    might require appropriate external preprocessing of filenames.

    Parameters
    ----------
    source : str or list
      file/folder path(s) to copy from the remote host
    destination : str
      file/folder path to copy to on the local host
    recursive : bool
      flag to enable recursive copying of given sources
    preserve_attrs : bool
      preserve modification times, access times, and modes from the
      original file

    Returns
    -------
    str
      stdout, stderr of the copy operation.
    """
    # make sure we have an open connection, will test if action is needed
    # by itself
    self.open()
    scp_cmd = self._get_scp_command_spec(recursive, preserve_attrs)
    # add source filepath(s) to scp command, prefixed with the remote host
    # PATCH in the line below: replaces `self.sshri.hostname` with `self.sshri.as_str()`
    scp_cmd += ["%s:%s" % (self.sshri.as_str(), self._quote_filename(s))
                for s in ensure_list(source)]
    # add destination path
    scp_cmd += [destination]
    out = self.runner.run(scp_cmd, protocol=StdOutErrCapture)
    return out['stdout'], out['stderr']


# The method 'BaseSSHConnection_put' is a patched version of
# 'datalad/support/sshconnector.py:BaseSSHConnection.put'
# from datalad@e0b357d9b8ca5f432638c23c0cb7c373028c8e52
def BaseSSHConnection_put(self, source, destination, recursive=False, preserve_attrs=False):
    """Copies source file/folder to destination on the remote.

    Note: this method performs escaping of filenames to an extent that
    moderately weird ones should work (spaces, quotes, pipes, other
    characters with special shell meaning), but more complicated cases
    might require appropriate external preprocessing of filenames.

    Parameters
    ----------
    source : str or list
      file/folder path(s) to copy from on local
    destination : str
      file/folder path to copy to on remote
    recursive : bool
      flag to enable recursive copying of given sources
    preserve_attrs : bool
      preserve modification times, access times, and modes from the
      original file

    Returns
    -------
    str
      stdout, stderr of the copy operation.
    """
    # make sure we have an open connection, will test if action is needed
    # by itself
    self.open()
    scp_cmd = self._get_scp_command_spec(recursive, preserve_attrs)
    # add source filepath(s) to scp command
    scp_cmd += ensure_list(source)
    # add destination path
    scp_cmd += ['%s:%s' % (
        # PATCH in the line below: replaces `self.sshri.hostname` with `self.sshri.as_str()`
        self.sshri.as_str(),
        self._quote_filename(destination),
    )]
    out = self.runner.run(scp_cmd, protocol=StdOutErrCapture)
    return out['stdout'], out['stderr']


apply_patch(
    modname='datalad.support.sshconnector',
    objname='BaseSSHConnection',
    attrname='get',
    patch=BaseSSHConnection_get,
)

apply_patch(
    modname='datalad.support.sshconnector',
    objname='BaseSSHConnection',
    attrname='put',
    patch=BaseSSHConnection_put,
)



"""Custom types and dataclasses

.. currentmodule:: datalad_next.types
.. autosummary::
   :toctree: generated

   AnnexKey
   ArchivistLocator
   ArchiveType
"""


from .annexkey import AnnexKey
from .archivist import ArchivistLocator
from .enums import ArchiveType
